(*         file: mgcky.ml
	         a CKY-like recognizer for MGs
        created: Thu Feb 28 09:12:22 PST 2008
     created by: E Stabler stabler@ucla.edu
        updated: 2010-05-08 11:21:21 PDT

   To compile: include grammar, remove #load unix, uncomment ui() in last line,
   then:
     ocamlc -o recognize unix.cma mgckyp.ml
   Or on platforms that have an optimized compiler, something like:
     ocamlopt.opt -inline 32 -unsafe -ccopt -O3 -o recognize unix.cmxa mgcky.ml
   (For runtime versions, print routines etc can be commented out)
*)

type feature = Sel of int | Cat of int | Pos of int | Neg of int;;
type chain = Em | Tu of (int * int * feature list);;  (* moving elements inside each item *)
type nmItem = chain array;;  (* non-movers, with cat coded by array address, have no features left *)
type mItem = int * int * feature list * chain array;; (* movers *)
type selItem = bool * feature list * chain array;;    (* bool in selItem says if item is lexical *)
type sitItem = int * int * bool * feature list * chain array;; (* situated items for agenda *)
type nmCatArray = Y | A of nmItem list array;;        (* length A arrays=(succ catMax) *)
type selArray = Z | B of selItem list array;;         (* length B arrays=(succ catMax) *)
type nmCatCoordChart = nmCatArray array array;;       (* n X n matrix, for n=succ (length input) *)
type selCoordChart = selArray array array;;           (* n X n matrix, for n=succ (length input) *)
type chart = nmCatCoordChart * selCoordChart * mItem list array;;

(* A grammar is needed to set size parameters moverMax, catMax, and specify categories.
   Just one grammar should be uncommented at a time!
#use "g-ne.ml";;   (* "titus praise -s lavinia" naive English SVIO, inspired by Mahajan 2000 *)
#use "g-nta.ml";;  (* "Beatrice Benedick criticize -s" naive Tamil SOVI, inspired by Mahajan 2000 *)
#use "g-nza.ml";;  (* "criticize -s Beatrice Benedick" naive Zapotec VISO, inspired by Mahajan 2000 *)
#use "tlingit.ml";; (* with a q-morpheme, inspired by Seth Cable *)
#use "anbn.ml";;   (* gets "a b" and "a a b b" but not "a" *)
#use "g4.ml";;     (* copy language aabaab *)
#use "g5.ml";;     (* generates 1^n2^n3^n4^n5^n as an s -- good for benchmarking *)
*)
#use "g-ne.ml";;

(* BEGIN print functions for tracing, and some other utilities *)
let fstOf3 (a,_,_) = a;;  (* so (fstOf3 chart) = nmCatCoordChart *)
let sndOf3 (_,a,_) = a;;  (* so (sndOf3 chart) = selCoordChart *)
let thdOf3 (_,_,a) = a;;  (* so (thdOf3 chart) = movers *)

let featString = function 
    Sel b -> "="^(catString b)
  | Cat b -> catString b
  | Pos b -> "+"^(moverString b)
  | Neg b -> "-"^(moverString b);;

let printFeat f = Printf.fprintf stdout "%s " (featString f);;
let printFeats fs = List.iter printFeat fs;;

let printStrings ss = 
  if ss=[]
  then Printf.fprintf stdout "[] " (* just to make the print presentation perfectly clear *)
  else List.iter (Printf.fprintf stdout "%s ") ss;;

let rec printLex grammar = match grammar with
    [] -> ()
  | (s,f)::xs -> 
      begin 
	printStrings s; Printf.fprintf stdout ":: ";
	printFeats f;  Printf.fprintf stdout "\n";
	printLex xs;
      end;;

let printChain simple = function 
    Em -> ()
  | Tu (i,j,feats) ->
      begin
	Printf.fprintf stdout "(%i,%i)" i j;
	if simple then Printf.fprintf stdout ":: " else Printf.fprintf stdout ": ";
	List.iter printFeat feats;
      end;;

let printChainWithF simple f = function 
    Em -> ()
  | Tu (i,j,feats) ->
      begin
	Printf.fprintf stdout "(%i,%i)" i j;
	if simple then Printf.fprintf stdout ":: " else Printf.fprintf stdout ": ";
	List.iter printFeat ((Neg f)::feats);
      end;;

(* to print nmItem, showing i,j coded into nXnNM array position, and Cat f coded into nmItem array position *)
let printNMItem i j f movers = 
  begin 
    printChain false (Tu (i,j,(Cat f)::[])); (* some non-movers might be lexical, but it does not matter *)
    Array.iteri (printChainWithF false) movers;  Printf.fprintf stdout "\n";
  end;;

let printNMItemList i j f itemlist = List.iter (printNMItem i j f) itemlist;;

let printNMCatArray i j array = match array with
    Y -> ()
  | A catArray -> Array.iteri (printNMItemList i j) catArray;;

let rec printNMCatCoordChart coordChart = 
  let n = pred (Array.length coordChart)
  in
    for i=0 to n do
      for j=0 to n do
	printNMCatArray i j coordChart.(i).(j)
      done
    done;;

let listArraySize array = Array.fold_right (function x -> (function y -> (List.length x) + y)) array 0;;

let nmCatArraySize catArray = match catArray with
    Y ->0
  | A array -> listArraySize array;;

let nmCatArrayArraySize catArray = 
  Array.fold_right (function x -> (function y -> (nmCatArraySize x) + y)) catArray 0;;

let nmChartSize catArray = 
  Array.fold_right (function x -> (function y -> (nmCatArrayArraySize x) + y)) catArray 0;;

(* to print selItem, showing i,j coded into nXnSel array position, and Sel f coded into selItem array position *)
let printSelItem i j f (simple,fs,movers) = 
  begin 
    printChain simple (Tu (i,j,(Sel f)::fs));
    Array.iteri (printChainWithF false) movers;  Printf.fprintf stdout "\n";
  end;;

let printSelItemList i j f itemlist = List.iter (printSelItem i j f) itemlist;;

let printSelArray i j array = match array with
    Z -> ()
  | B catArray -> Array.iteri (printSelItemList i j) catArray;;

let rec printSelCoordChart coordChart = 
  let n = pred (Array.length coordChart)
  in
    for i=0 to n do
      for j=0 to n do
	printSelArray i j coordChart.(i).(j)
      done
    done;;

let selArraySize catArray = match catArray with
    Z -> 0
  | B array -> listArraySize array;;

let selArrayArraySize catArray = 
  Array.fold_right (function x -> (function y -> (selArraySize x) + y)) catArray 0;;

let selChartSize catArray = 
  Array.fold_right (function x -> (function y -> (selArrayArraySize x) + y)) catArray 0;;

(* to print sitItem in moverCats, insert Cat f that is coded into nmItem array position *)
let printMoverItem f (i,j,fs,movers) = 
  begin 
    printChain false (Tu (i,j,(Cat f)::fs)); (* we say "false", but it does not matter *)
    Array.iteri (printChainWithF false) movers;  Printf.fprintf stdout "\n";
  end;;

let printMoverItemList f itemlist = List.iter (printMoverItem f) itemlist;;

let printMoverArray = Array.iteri printMoverItemList;;

(* to print sitItem in agenda *)
let printOtherItem (i,j,simple,fs,movers) = 
  begin 
    printChain simple (Tu (i,j,fs));
    Array.iteri (printChainWithF false) movers;  Printf.fprintf stdout "\n";
  end;;

let printOtherItemList itemlist = List.iter printOtherItem itemlist;;

let printChart (nXnNonmoverCat, nXnSel, moverCats) =
  begin
    printNMCatCoordChart nXnNonmoverCat;
    printSelCoordChart nXnSel;
    printMoverArray moverCats;
  end;;

let chartStats (nXnNonmoverCat, nXnSel, moverCats) =
  let nmSize = nmChartSize nXnNonmoverCat in
  let selSize = selChartSize nXnSel in
  let moverSize = listArraySize moverCats in
  let total = nmSize+selSize+moverSize in
    Printf.fprintf stdout "Chart size (%i,%i,%i) = %i total," nmSize selSize moverSize total;;

let rec printAgenda agenda = 
  begin
    Printf.fprintf stdout "---agenda:\n";
    printOtherItemList agenda;
    Printf.fprintf stdout "---end agenda\n";
  end;;
(** END print functions for tracing **)

let rec ensureMember e list = match list with
    [] -> [e]
  | x::xs -> if e=x then x::xs else x::ensureMember e xs;;

(* put empties into *every string position* by calling: emptyInsert lex 0 inputLength (agenda0,chart0) lex. *)
let rec emptyInsert lex i stop (agenda,(nXnNM,nXnSel,movers)) remaining = match remaining with
    [] -> 
      if i = stop
      then agenda
      else emptyInsert lex (succ i) stop (agenda,(nXnNM,nXnSel,movers)) lex
  | ([],(Cat x)::[])::moreLex ->  (* empty non-moving Cat -- i.e. just 1 feature! *)
      begin
	(match nXnNM.(i).(i) with
	    Y -> let catArray = Array.make (succ catMax) [] 
	    in 
	      begin
		catArray.(x) <- (Array.make (succ moverMax) Em :: []); 
		nXnNM.(i).(i) <- A catArray;
	      end
	  | A catArray -> 
	      begin
		catArray.(x) <- (Array.make (succ moverMax) Em :: catArray.(x));
	      end
	);
	emptyInsert lex i stop ((i,i,true,[Cat x],(Array.make (succ moverMax) Em))::agenda,(nXnNM,nXnSel,movers)) moreLex;
      end
  | ([],(Cat x)::fs)::moreLex ->  (* empty moving Cats *)
      begin
	if i=0 
	then 
	  begin
	    movers.(x) <- (i,i,fs,Array.make (succ moverMax) Em)::movers.(x);
	    emptyInsert lex i stop ((i,i,true,(Cat x)::fs,Array.make (succ moverMax) Em)::agenda,(nXnNM,nXnSel,movers)) moreLex;
	  end
	else emptyInsert lex i stop (agenda,(nXnNM,nXnSel,movers)) moreLex
      end
  | ([],(Sel x)::fs)::moreLex ->  (* empty selector, add to chart **and to agenda** *)
      begin
	(match nXnSel.(i).(i) with
	    Z -> let catArray = Array.make (succ catMax) [] 
	    in 
	      begin
		catArray.(x) <- ((true, fs, Array.make (succ moverMax) Em) :: []);
		nXnSel.(i).(i) <- B catArray;
	      end
	  | B catArray -> 
	      begin
		catArray.(x) <- ((true, fs, (Array.make (succ moverMax) Em)) :: catArray.(x));
	      end
	);
	  emptyInsert lex i stop ((i,i,true,(Sel x)::fs,Array.make (succ moverMax) Em)::agenda ,(nXnNM,nXnSel,movers)) moreLex;
      end
  | _::moreLex -> emptyInsert lex i stop (agenda,(nXnNM,nXnSel,movers)) moreLex;;

(* add non-empty words to the chart *)
let rec lexInsert lex (agenda,(nXnNM,nXnSel,movers)) i input remainingG = match (input,remainingG) with
    ([],_) -> agenda
  | (w::ws,[]) -> lexInsert lex (agenda,(nXnNM,nXnSel,movers)) (succ i) ws lex
  | (w::ws,([word],(Cat x)::[])::moreLex) when w=word ->   (* non-moving Cat -- i.e. just 1 feature! *)
      begin
	(match nXnNM.(i).(succ i) with
	    Y -> let catArray = Array.make (succ catMax) [] 
	    in 
	      begin
		catArray.(x) <- (Array.make (succ moverMax) Em :: []);
		nXnNM.(i).(succ i) <- A catArray;
	      end
	  | A catArray -> 
	      begin
		catArray.(x) <- (Array.make (succ moverMax) Em :: catArray.(x));
	      end
	);
	lexInsert lex ((i,succ i,true,[Cat x],(Array.make (succ moverMax) Em))::agenda,(nXnNM,nXnSel,movers)) i input moreLex;
      end	  
  | (w::ws,([word],(Cat x)::fs)::moreLex) when w=word ->  let cell = movers.(x) in  (* moving Cat *)
      if List.mem (i,succ i,fs,Array.make (succ moverMax) Em) cell
      then lexInsert lex (agenda,(nXnNM,nXnSel,movers)) i input moreLex
      else 
	begin 
	  movers.(x) <- (i,succ i,fs,Array.make (succ moverMax) Em)::cell;
	  lexInsert lex ((i,succ i,true,(Cat x)::fs,Array.make (succ moverMax) Em)::agenda,(nXnNM,nXnSel,movers)) i input moreLex;
	end
  | (w::ws,([word],(Sel x)::fs)::moreLex) when w=word ->  (* selector, add to chart **and to agenda** *)
      begin
	(match nXnSel.(i).(succ i) with
	    Z -> let catArray = Array.make (succ catMax) [] 
	    in 
	      begin
		catArray.(x) <- ((true, fs, Array.make (succ moverMax) Em) :: []);
		nXnSel.(i).(succ i) <- B catArray;
	      end
	  | B catArray -> 
	      begin
		catArray.(x) <- ((true, fs, (Array.make (succ moverMax) Em)) :: catArray.(x));
	      end
	);
	let newAgenda = (i,succ i,true,(Sel x)::fs,Array.make (succ moverMax) Em)::agenda in 
	  lexInsert lex (newAgenda,(nXnNM,nXnSel,movers)) i input moreLex;
      end
  | (w::ws,_::moreLex) -> lexInsert lex (agenda,(nXnNM,nXnSel,movers)) i input moreLex;;

(** NOW UTILITIES FOR MAIN LOOP, COMPUTING THE CLOSURE **)
(* combine the moving elements of two constituents, if possible, returning (newArray,success) *)
let rec moverMergeVal array1 array2 newArray i stop =
  if i>=stop
  then (newArray,true)
  else if array1.(i)=Em then begin newArray.(i)<-array2.(i); moverMergeVal array1 array2 newArray (succ i) stop; end
  else if array2.(i)=Em then begin newArray.(i)<-array1.(i); moverMergeVal array1 array2 newArray (succ i) stop; end
  else (array1,false);;

(* update (agenda,chart) with each possible kind of item, after applying move if necessary *)
(* update needs to be recursive because of the possibility of arbitrary numbers of move steps *)
let rec update (i,j,simple,feats,moverArray) (agenda,(nXnNM,nXnSel,movers)) = 
  let rec move (k,l,x,features,moverArray) agendaChart =                      (* that's the letter l *)
      (match moverArray.(x) with
	 Em -> agendaChart
       | Tu (i,j,[]) -> (* move1: final move *)
	   if j=k
	   then let newMoverArray = Array.copy moverArray in
	     begin
	       newMoverArray.(x) <- Em;
	       update (i,l,false,features,newMoverArray) agendaChart;
	     end
	   else agendaChart
       | Tu (i,j,(Neg f)::fs) -> (* move2: nonfinal *)
	   if moverArray.(f) = Em
	   then let newMoverArray = Array.copy moverArray in
	     begin
	       newMoverArray.(x) <- Em;
	       newMoverArray.(f) <- Tu (i,j,fs);
	       update (k,l,false,features,newMoverArray) agendaChart;
	     end
	   else agendaChart
       | Tu _ -> failwith "move error: invalid licensee sequence"
      )
  in match feats with
      [] -> failwith "update error: featureless result"
    | (Cat x) :: [] -> (* nonmoving Cat item *)
	(match nXnNM.(i).(j) with
	     Y -> let catArray = Array.make (succ catMax) [] in 
	       begin
		 catArray.(x) <- (moverArray :: []);
		 nXnNM.(i).(j) <- A catArray;
		 ((i,j,false,(Cat x)::[],moverArray)::agenda,(nXnNM,nXnSel,movers));
	       end
	   | A catArray -> 
	       let cell = catArray.(x) in
		 if List.mem moverArray cell
		 then (agenda,(nXnNM,nXnSel,movers))
		 else 
		   begin
(*		     Printf.fprintf stdout "adding: "; printOtherItem (i,j,false,(Cat x)::[],moverArray); *)
		     catArray.(x) <- (moverArray::cell);
		     ((i,j,false,(Cat x)::[],moverArray)::agenda,(nXnNM,nXnSel,movers));
		   end
	)
    | (Cat x) :: features -> (* moving Cat item *)
	let cell = movers.(x) in
	  if List.mem (i,j,features,moverArray) cell
	  then (agenda,(nXnNM,nXnSel,movers))
	  else 
	    begin
(*	      Printf.fprintf stdout "adding: "; printOtherItem (i,j,false,(Cat x)::features,moverArray); *)
	      movers.(x) <- (i,j,features,moverArray)::cell;
	      ((i,j,false,(Cat x)::features,moverArray)::agenda,(nXnNM,nXnSel,movers));
	    end
    | (Sel x) :: features ->
	(match nXnSel.(i).(j) with
	     Z -> let catArray = Array.make (succ catMax) [] 
	     in 
	       begin
(*		 Printf.fprintf stdout "adding: "; printOtherItem (i,j,false,(Sel x)::features,moverArray); *)
		 catArray.(x) <- ((false, features, moverArray) :: []);
		 nXnSel.(i).(j) <- B catArray;
		 ((i,j,false,(Sel x)::features,moverArray)::agenda,(nXnNM,nXnSel,movers));
	       end
	   | B catArray -> 
	       let cell = catArray.(x) in
		 if List.mem (false, features, moverArray) cell
		 then (agenda,(nXnNM,nXnSel,movers))
		 else 
		   begin
(*		     Printf.fprintf stdout "adding: "; printOtherItem (i,j,false,(Sel x)::features,moverArray); *)
		     catArray.(x) <- ((false, features, moverArray) :: catArray.(x));
		     ((i,j,false,(Sel x)::features,moverArray)::agenda,(nXnNM,nXnSel,movers));
		   end
	)
    | (Pos x) :: features ->
(*	Printf.fprintf stdout "trying move: "; printOtherItem (i,j,false,(Pos x)::features,moverArray); *)
	move (i,j,x,features,moverArray) (agenda,(nXnNM,nXnSel,movers))
    | (Neg x) :: features -> failwith "update error: Neg result";;

(* for forward rules merge1, merge2, merge3, this function calls update on all matches with Cat items *)
let rec listAll (agenda,chart) (i,k,fs,movers1) cell = match cell with
    [] -> agenda;
  | movers2::more ->
      let result = moverMergeVal movers1 movers2 (Array.make (succ moverMax) Em) 0 (succ moverMax) in  
	if (snd result)
	then listAll (update (i,k,false,fs,(fst result)) (agenda,chart)) (i,k,fs,movers1) more
	else listAll (agenda,chart) (i,k,fs,movers1) more;;

(* MERGE1: search nXnNM(j)(k) up to nXnNM(j)(stop) for matches with agenda item (pos1,i,j,true,(Sel x)::fs,movers1) *)
let rec getall1 (agenda,(nXnNM,nXnSel,movers)) (i,fs,movers1) x j k stop = 
  if k > stop
  then agenda
  else match nXnNM.(j).(k) with
      Y -> getall1 (agenda,(nXnNM,nXnSel,movers)) (i,fs,movers1) x j (succ k) stop
    | A array ->
	  let newAgenda = listAll (agenda,(nXnNM,nXnSel,movers)) (i,k,fs,movers1) array.(x) in
	    getall1 (newAgenda,(nXnNM,nXnSel,movers)) (i,fs,movers1) x j (succ k) stop;;

(* MERGE2: search nXnNM(i)(j) down to nXnNM(0)(j) for matches with agenda item (pos1,j,k,false,(Sel x)::fs,movers1) *)
let rec getall2 (agenda,(nXnNM,nXnSel,movers)) (k,fs,movers1) x i j = 
  if i < 0
  then agenda
  else match nXnNM.(i).(j) with
      Y -> getall2 (agenda,(nXnNM,nXnSel,movers)) (k,fs,movers1) x (pred i) j
    | A array -> 
	let newAgenda = listAll (agenda,(nXnNM,nXnSel,movers)) (i,k,fs,movers1) array.(x) in
	  getall2 (newAgenda,(nXnNM,nXnSel,movers)) (k,fs,movers1) x (pred i) j;;

(* MERGE3  search for matches with agenda item (pos1,i,j,true/false,(Sel x)::fs,movers1) *)
let rec getall3 (agenda,chart) (i,j,fs1,movers1) cell = match cell with
    [] -> agenda
  | (k,l,(Neg y)::fs2,movers2)::more ->
      if movers1.(y) = Em && movers2.(y)=Em
      then 
	  let (newMovers,ok) = moverMergeVal movers1 movers2 (Array.make (succ moverMax) Em) 0 (succ moverMax) in
	    if ok
	    then
	      begin
 		newMovers.(y) <- Tu (k,l,fs2);
		getall3 (update (i,j,false,fs1,newMovers) (agenda,chart)) (i,j,fs1,movers1) more;
	      end
	    else getall3 (agenda,chart) (i,j,fs1,movers1) more
      else getall3 (agenda,chart) (i,j,fs1,movers1) more
  | _ -> failwith "getall3: error";;

(* for backward rules merge1r, merge2r, this function calls update on all matches with Sel items *)
(* The cell we are looking at here is, in effect, nXnSel.(i).(j).(x), so only i,k are passed *)
let rec listAllr (agenda,chart) (i,k,movers1) cell = match cell with
    [] -> agenda
  | (simple,fs,movers2)::more -> 
	let result = moverMergeVal movers1 movers2 (Array.make (succ moverMax) Em) 0 (succ moverMax) in  
	  if (snd result)
	  then listAllr (update (i,k,false,fs,(fst result)) (agenda,chart)) (i,k,movers1) more
	  else listAllr (agenda,chart) (i,k,movers1) more;;

(* MERGE1r: search nXnSel(i)(j) down to nXnSel(0)(j) for matches with agenda item (pos1,j,k,_,(Cat x)::[],movers1) *)
let rec getall1r (agenda,(nXnNM,nXnSel,movers)) (k,movers1) x i j = 
  if i < 0
  then agenda
  else match nXnSel.(i).(j) with
      Z -> getall1r (agenda,(nXnNM,nXnSel,movers)) (k,movers1) x (pred i) j
    | B array -> 
	let newAgenda = listAllr (agenda,(nXnNM,nXnSel,movers)) (i,k,movers1) array.(x) in
	  getall1r (newAgenda,(nXnNM,nXnSel,movers)) (k,movers1) x (pred i) j;;

(* MERGE2r: search nXnNM(j)(k) up to nXnNM(j)(stop) for matches with agenda item (pos1,j,k,_,(Cat x)::[],movers1) *)
let rec getall2r (agenda,(nXnNM,nXnSel,movers)) (i,movers1) x j k stop = 
  if k > stop
  then agenda
  else match nXnSel.(j).(k) with
      Z -> getall2r (agenda,(nXnNM,nXnSel,movers)) (i,movers1) x j (succ k) stop
    | B array -> 
	let newAgenda = listAllr (agenda,(nXnNM,nXnSel,movers)) (i,k,movers1) array.(x) in
	      getall2r (newAgenda,(nXnNM,nXnSel,movers)) (i,movers1) x j (succ k) stop;;

(* for backward rule merge3r, this function calls update on all matches with Sel items *)
(* MERGE3r: category x of moving i,j matches all selector x at k,l in nXnSel(k)(l) *)
let rec listAll3r (agenda,chart) (i,j,k,l,fs1,movers1) cell = match cell with
    [] -> agenda
  | (simple,fs2,movers2)::more -> (* first, make sure that movers1 and movers2 are compatible *)
      let result = moverMergeVal movers1 movers2 (Array.make (succ moverMax) Em) 0 (succ moverMax) in  
	  if (snd result)
	  then (match fs1 with
		    (Neg y)::morefs1 ->	    (* then, put new moving feature into result, if compatible *)
		      if (fst result).(y) = Em
		      then
			begin
			  (fst result).(y) <- Tu (i,j,morefs1);
			  listAll3r (update (k,l,false,fs2,(fst result)) (agenda,chart)) (i,j,k,l,fs1,movers1) more;
			end
		      else listAll3r (agenda,chart) (i,j,k,l,fs1,movers1) more
		| _ -> failwith "listAll3r: invalid feature in licensee sequence"
	       )
	  else listAll3r (agenda,chart) (i,j,k,l,fs1,movers1) more;;

(* MERGE3r: category x of moving i,j finds selector x of k,l in nXnSel(k)(l), for any (non-overlapping) k,l *)
(* The k<=l restriction to the upper triangle of the matrix is coded by going to (succ k) (succ k) when l > stop *)
let rec getall3r (agenda,(nXnNM,nXnSel,movers)) (i,j,fs,movers1) x k l stop =   (* letter l *)
  if l > stop && k >= stop then agenda
  else if l > stop then getall3r (agenda,(nXnNM,nXnSel,movers)) (i,j,fs,movers1) x (succ k) (succ k) stop
  else if (l <= i || k >= j)
  then 
      (match nXnSel.(k).(l) with
	   Z -> getall3r (agenda,(nXnNM,nXnSel,movers)) (i,j,fs,movers1) x k (succ l) stop
	 | B array -> 
	     let newAgenda = listAll3r (agenda,(nXnNM,nXnSel,movers)) (i,j,k,l,fs,movers1) array.(x) in
	       getall3r (newAgenda,(nXnNM,nXnSel,movers)) (i,j,fs,movers1) x k (succ l) stop
      )
  else getall3r (agenda,(nXnNM,nXnSel,movers)) (i,j,fs,movers1) x k (succ l) stop;;

(* MAIN LOOP TO COMPUTE THE CLOSURE OF THE CHART *)
let rec main stop (agenda,chart) = 
  begin
    (* printAgenda agenda;    for tracing, but often produces a lot of output! *)
    (* printChart chart;      for tracing, but often produces a lot of output! *)
    (match agenda with
	 [] -> ()
       | (i,j,true,(Sel x)::fs,movers1)::moreAgenda ->  (* merge1, merge3 *)
	   let midAgenda = getall1 (moreAgenda,chart) (i,fs,movers1) x j j stop in
	   let newAgenda = getall3 (midAgenda,chart) (i,j,fs,movers1) (thdOf3 chart).(x) in 
	     main stop (newAgenda,chart)
       | (j,k,false,(Sel x)::fs,movers1)::moreAgenda ->   (* merge2, merge3 *)
	   let midAgenda = getall2 (moreAgenda,chart) (k,fs,movers1) x j j in
	   let newAgenda = getall3 (midAgenda,chart) (j,k,fs,movers1) (thdOf3 chart).(x) in
	     main stop (newAgenda,chart)
       | (j,k,_,(Cat x)::[],movers1)::moreAgenda ->  (* merge1r, merge2r *)
	   let midAgenda = getall1r (moreAgenda,chart) (k,movers1) x j j in
	   let newAgenda = getall2r (midAgenda,chart) (j,movers1) x k k stop in
	     main stop (newAgenda,chart)
       | (i,j,_,(Cat x)::fs,movers1)::moreAgenda ->  (* merge3r *)
	   let newAgenda = getall3r (moreAgenda,chart) (i,j,fs,movers1) x 0 0 stop in
	     main stop (newAgenda,chart)
       | _ -> failwith "main: invalid agenda item"
    );
  end;;

let rec emptyChain chain i stop =  (* true if chain has no moving elements in it *)
  if i>=stop then true
  else if chain.(i)=Em then emptyChain chain (succ i) stop
  else false;;

let rec analyzeNMitems inputLength startCategory cell = match cell with
    [] -> Printf.fprintf stdout "No consitutent of category %s from %i to %i.\n" (catString startCategory) 0 inputLength
  | chain::chains ->
      if emptyChain chain 0 (Array.length chain)
      then Printf.fprintf stdout "There is a consitutent of category %s from %i to %i.\n" (catString startCategory) 0 inputLength
      else analyzeNMitems inputLength startCategory chains;;

let analyze nXnNM startCategory inputLength = match nXnNM.(0).(inputLength) with
    Y -> Printf.fprintf stdout "No consitutent of category %s from %i to %i.\n" (catString startCategory) 0 inputLength
  | A array -> analyzeNMitems inputLength startCategory array.(startCategory);;

(* *)
#load "unix.cma";;  (* for getting cpu time *)

let cky lex (startCat,input) = 
  let time0 = (Unix.times()).Unix.tms_utime in
  let inputLength0 = List.length input in
  let (chart0:chart) = 
    ( Array.make_matrix (succ inputLength0) (succ inputLength0) Y,
      Array.make_matrix (succ inputLength0) (succ inputLength0) Z,
      Array.make (succ catMax) []
    ) in
  let fstAgenda = emptyInsert lex 0 inputLength0 ([],chart0) lex in
  let sndAgenda = lexInsert lex (fstAgenda,chart0) 0 input lex in
    begin
      main inputLength0 (sndAgenda,chart0);
      let time1 = (Unix.times()).Unix.tms_utime in
	begin
(*	  printChart chart0; *)
	  chartStats chart0; Printf.printf " computed in %2.2f ms.\n" ((time1 -. time0));
	  analyze (fstOf3 chart0) startCat inputLength0;
	end;
    end;;

let eg g x = cky g (example x);;

(* eg grammar 2;; *)

(* use to test compiled code 
let rec uiLoop () = 
  begin
    Printf.fprintf stdout "example (0 for stop): ";
    let next = (read_line ()) in 
      if next.[0] = '0' 
      then Printf.fprintf stdout "Happy to be of service. Ciao.\n"
      else
	begin
	  eg grammar (int_of_string next);
	  uiLoop ();
	end
  end;;

uiLoop ();;
*)
